from __future__ import annotations

import re
from typing import TYPE_CHECKING, Optional, cast

from rich.console import RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.measure import measure_renderables
from rich.pretty import Pretty
from rich.protocol import is_renderable
from rich.segment import Segment
from rich.text import Text
from textual.cache import LRUCache
from textual.geometry import Region, Size
from textual.reactive import var
from textual.scroll_view import ScrollView
from textual.strip import Strip

if TYPE_CHECKING:
    from typing_extensions import Self


class LogLines(ScrollView, can_focus=True):
    COMPONENT_CLASSES = {
        "loglines--filter-highlight",
        "loglines--filter-highlight-selected",
    }

    max_lines: var[int | None] = var[Optional[int]](None)
    min_width: var[int] = var(78)
    wrap: var[bool] = var(False)
    highlight: var[bool] = var(False)
    markup: var[bool] = var(False)
    auto_scroll: var[bool] = var(True)

    def __init__(
        self,
        *,
        max_lines: int | None = None,
        min_width: int = 78,
        wrap: bool = False,
        highlight: bool = False,
        markup: bool = False,
        auto_scroll: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.max_lines = max_lines
        self.keyword = None
        self.current_index = None
        self.case_sensitive = False
        self._start_line: int = 0
        self.lines: list[Strip] = []
        self._line_cache: LRUCache[tuple[int, int, int, int], Strip]
        self._line_cache = LRUCache(1024)
        self.max_width: int = 0
        self.min_width = min_width
        self.wrap = wrap
        self.highlight = highlight
        self.markup = markup
        self.auto_scroll = auto_scroll
        self.highlighter: Highlighter = ReprHighlighter()
        self._last_container_width: int = min_width

    def notify_style_update(self) -> None:
        self._line_cache.clear()

    def on_resize(self) -> None:
        self._last_container_width = self.scrollable_content_region.width

    def _make_renderable(self, content: RenderableType | object) -> RenderableType:
        renderable: RenderableType
        if not is_renderable(content):
            renderable = Pretty(content)
        else:
            if isinstance(content, str):
                if self.markup:
                    renderable = Text.from_markup(content)
                else:
                    renderable = Text(content)
                if self.highlight:
                    renderable = self.highlighter(renderable)
            else:
                renderable = cast(RenderableType, content)

        if isinstance(renderable, Text):
            renderable.expand_tabs()

        return renderable

    def write(
        self,
        content: RenderableType | object,
        width: int | None = None,
        expand: bool = False,
        shrink: bool = True,
        scroll_end: bool | None = None,
    ) -> Self:
        auto_scroll = self.auto_scroll if scroll_end is None else scroll_end

        console = self.app.console
        render_options = console.options

        renderable = self._make_renderable(content)

        if isinstance(renderable, Text) and not self.wrap:
            render_options = render_options.update(overflow="ignore", no_wrap=True)

        render_width = measure_renderables(
            console, render_options, [renderable]
        ).maximum

        container_width = (
            self.scrollable_content_region.width if width is None else width
        )

        container_width = (
            container_width if container_width else self._last_container_width
        )

        if expand and render_width < container_width:
            render_width = container_width
        if shrink and render_width > container_width:
            render_width = container_width

        render_width = max(render_width, self.min_width)

        segments = self.app.console.render(
            renderable, render_options.update_width(render_width)
        )
        lines = list(Segment.split_lines(segments))
        if not lines:
            self.lines.append(Strip.blank(render_width))
        else:
            self.max_width = max(
                self.max_width,
                max(sum([segment.cell_length for segment in _line]) for _line in lines),
            )
            strips = Strip.from_lines(lines)
            for strip in strips:
                strip.adjust_cell_length(render_width)
            self.lines.extend(strips)

            if self.max_lines is not None and len(self.lines) > self.max_lines:
                self._start_line += len(self.lines) - self.max_lines
                self.refresh()
                self.lines = self.lines[-self.max_lines :]
        self.virtual_size = Size(self.max_width, len(self.lines))
        if auto_scroll:
            self.scroll_end(animate=False)

        return self

    def clear(self) -> Self:
        self.lines.clear()
        self._line_cache.clear()
        self._start_line = 0
        self.max_width = 0
        self.virtual_size = Size(self.max_width, len(self.lines))
        self.refresh()
        return self

    def render_line(self, y: int) -> Strip:
        filter_style_line = self.get_component_rich_style("loglines--filter-highlight")
        filter_style_word = self.get_component_rich_style(
            "loglines--filter-highlight-selected"
        )
        scroll_x, scroll_y = self.scroll_offset
        line = self._render_line(scroll_y + y, scroll_x, self.size.width)
        text = Text.from_ansi(line.text)

        if self.keyword:
            matches = list(
                re.finditer(
                    self.keyword,
                    line.text,
                    flags=self.case_sensitive,
                )
            )

            if self.current_index == scroll_y + y:
                text.stylize(filter_style_line, 0, text.cell_len)

            if matches:
                for match in matches:
                    text.stylize(filter_style_word, *match.span())
            else:
                text.stylize("dim")

            strip = Strip(text.render(self.app.console), text.cell_len)

        else:
            strip = line.apply_style(self.rich_style)

        return strip

    def render_lines(self, crop: Region) -> list[Strip]:
        lines = self._styles_cache.render_widget(self, crop)
        return lines

    def _render_line(self, y: int, scroll_x: int, width: int) -> Strip:
        if y >= len(self.lines):
            return Strip.blank(width, self.rich_style)

        key = (y + self._start_line, scroll_x, width, self.max_width)
        if key in self._line_cache:
            return self._line_cache[key]

        line = self.lines[y].crop_extend(scroll_x, scroll_x + width, self.rich_style)

        self._line_cache[key] = line
        return line



import webbrowser

import requests
from rich.text import Text
from textual import on
from textual.app import ComposeResult
from textual.containers import Center, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Markdown, Static

from application.util.helper import POCKER_CONFIG_BASE_PATH, get_current_version

HELP_MD = """
Pocker is a tool for the terminal to do Docker related tasks.

Repository: [https://github.com/pommee/Pocker](https://github.com/pommee/Pocker)  
Author: [pommee](https://github.com/pommee)

---

- `esc` closes this window.

### Navigation

- `q` to quit pocker.
- `?` shows this help modal.
- `l` updates the content window with logs of selected container.
- `/` search in the content window.
- `a` show container attributes.
- `e` show container environment variables.
- `f` display content window in full-size.
- `v` shell in the container.
- `w` will wrap logs/attributes to avoid horizontal scrolling.

### Other keys

- `ctrl+f` or `/` Show find dialog.

"""

TITLE = rf"""
  _____   _____  _______ _     _ _______  ______
 |_____] |     | |       |____/  |______ |_____/
 |       |_____| |_____  |    \_ |______ |    \_  v{get_current_version()}

"""


def get_title() -> Text:
    lines = TITLE.splitlines(keepends=True)
    return Text.assemble(
        *zip(
            lines,
            [
                "#6076FF",
                "#0087FF",
                "#00888B",
                "#008157",
            ],
        )
    )


class HelpScreen(ModalScreen):
    BINDINGS = [
        ("escape", "dismiss"),
    ]

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            with Center():
                yield Static(get_title(), id="title", classes="title")
            yield Markdown(HELP_MD, id="help", classes="help")
            yield Markdown(self.read_changelog(), id="changelog", classes="changelog")

    @on(Markdown.LinkClicked)
    def on_markdown_link_clicked(self, event: Markdown.LinkClicked) -> None:
        self.action_go(event.href)

    def action_go(self, href: str) -> None:
        webbrowser.open(href)

    def read_changelog(self):
        file_path = POCKER_CONFIG_BASE_PATH / "CHANGELOG.md"
        if not file_path.exists():
            response = requests.get(
                "https://raw.githubusercontent.com/pommee/Pocker/main/CHANGELOG.md"
            )
            if response.status_code == 200:
                with open(file_path, "w") as file:
                    file.write(response.text)
            else:
                return response.text

        with open(file_path, "r") as file:
            return "---  \n# Changelog  \n\n" + file.read()



from dataclasses import dataclass

from docker.models.images import Image
from textual import on
from textual.app import ComposeResult
from textual.message import Message
from textual.widget import Widget
from textual.widgets import (
    Label,
    ListItem,
    ListView,
)

from application.docker_manager import DockerManager


class PockerImages(Widget):
    BORDER_TITLE: str = "Images"

    def __init__(
        self,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.docker_manager = docker_manager
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        with ListView(id="ContainersAndImagesListView"):
            image: Image
            for image in self.docker_manager.images:
                if image.tags:
                    name = image.tags[0].split(":")[0]
                    version = image.tags[0].split(":")[1]
                    yield ListItem(Label(f"{name}:{version}"))

    @dataclass
    class ClickedImage(Message, bubble=True):
        clicked_image: ListItem

    # TODO: Implement fetching relevant information regarding an image.
    def on_list_view_selected(self, selected: ListView.Selected):
        self.notify(
            title="Not implemented",
            message="Viewing images will be supported in future releases.",
            timeout=6,
        )
        # self.post_message(PockerImages.ClickedImage(selected.item))

    @on(ClickedImage)
    def _update_selected_container(self, new_selected_image: ClickedImage):
        selected_image = new_selected_image.clicked_image
        images = self.query_one(ListView)
        image: ListItem
        for image in images.children:
            if image.has_class("selected"):
                image.remove_class("selected")

        selected_image.add_class("selected")



import os

import subprocess
from threading import Thread

if os.name == "nt":
    from winpty import PTY
else:
    import pty

from rich.text import Text, TextType
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import (
    Input,
    RichLog,
    TabPane,
)

from application.docker_manager import DockerManager


class ShellPane(TabPane):
    def __init__(
        self,
        title: TextType,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        self.docker_manager = docker_manager
        super().__init__(
            title, *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        yield RichLog(
            id="shell-output",
            highlight=True,
            auto_scroll=True,
            name="ShellPaneLog",
        )
        yield Input(placeholder="Enter command...", id="shell-input")

    def run_shell(self) -> None:
        self.container = self.docker_manager.selected_container
        self.output_widget = self.query_one("#shell-output", RichLog)
        self.input_widget = self.query_one("#shell-input", Input)
        self.default_shell = "bash"

        if os.name == "nt":
            self.process = PTY(80, 25)
            if not self.process.spawn(
                f"docker exec -it {self.container.id} {self.default_shell}"
            ):
                raise Exception("Failed to spawn shell")
        else:
            self.master_fd, slave_fd = pty.openpty()
            self.process = subprocess.Popen(
                ["docker", "exec", "-it", self.container.id, self.default_shell],
                stdin=slave_fd,
                stdout=slave_fd,
                stderr=slave_fd,
                preexec_fn=os.setsid,
            )

        def read_output(fd):
            while True:
                if os.name == "nt":
                    output = fd.read(16384)
                    if not output:
                        continue
                else:
                    output = os.read(fd, 16384).decode()
                self.app.call_from_thread(self.write_output, output)

        if os.name == "nt":
            self.read_thread = Thread(target=read_output, args=(self.process,))
        else:
            self.read_thread = Thread(target=read_output, args=(self.master_fd,))
        self.read_thread.daemon = True
        self.read_thread.start()

    def write_output(self, output: str) -> None:
        self.output_widget.write(Text.from_ansi(output))

    async def on_input_submitted(self, message: Input.Submitted) -> None:
        self.send_command(message.value)
        self.input_widget.clear()

    def send_command(self, command: str) -> None:
        if self.process:
            if os.name == "nt":
                self.process.write(command + "\n")
            else:
                os.write(self.master_fd, (command + "\n").encode())



import re
from threading import Event, Thread
from typing import Any

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import (
    Input,
    Static,
    Switch,
    TabbedContent,
    TabPane,
)

from application.docker_manager import DockerManager
from application.widget.log_viewer import LogLines
from application.widget.shell import ShellPane


class ContentWindow(Widget):
    current_list_index = None
    search_keyword = ""
    matches, indices = None, None
    log_task_stop_event = Event()

    def __init__(
        self,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.docker_manager = docker_manager
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        global logs

        yield Horizontal(
            Input(id="search_log_input", placeholder="Search logs...", type="text"),
            Horizontal(
                Static("case-sensitive", classes="label"),
                Switch(id="case-sensitive-switch", animate=True),
                classes="case-sensitive-switch",
            ),
            classes="container",
        )
        with TabbedContent():
            with TabPane("Logs", id="logpane"):
                logs = LogLines(id="logs", highlight=True, auto_scroll=True, name="Log")
                logs.border_title = self.docker_manager.selected_container.name
                logs.scroll_end(animate=False)
                yield logs
            with TabPane("Attributes", id="attributespane"):
                attributes = LogLines(
                    id="attributes_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Attributes",
                )
                attributes.border_title = self.docker_manager.selected_container.name
                attributes.scroll_end(animate=False)
                yield attributes
            with TabPane("Environment", id="environmentpane"):
                environment = LogLines(
                    id="environment_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Environment",
                )
                environment.border_title = self.docker_manager.selected_container.name
                environment.scroll_end(animate=False)
                yield environment
            with TabPane("Statistics", id="statisticspane"):
                statistics = LogLines(
                    id="statistics_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Statistics",
                )
                statistics.border_title = self.docker_manager.selected_container.name
                statistics.scroll_end(animate=False)
                yield statistics
            yield ShellPane(
                title="Shell", id="shellpane", docker_manager=self.docker_manager
            )

    def search_logs(self, pattern):
        case_sensitive_switch = self.query_one("#case-sensitive-switch", Switch).value
        case_sensitive = 0 if case_sensitive_switch else re.IGNORECASE
        compiled_pattern = re.compile(pattern, case_sensitive)

        current_tab = self.query_one(TabbedContent).active
        self.current_content_window: LogLines = self.query_one(
            f"#{current_tab}"
        ).children[0]
        results = [
            (line, i)
            for i, line in enumerate(self.current_content_window.lines)
            if compiled_pattern.search(line.text)
        ]

        if results:
            self.matches, self.indices = zip(*results)
            self.matches = list(self.matches)
            self.indices = list(self.indices)
        else:
            self.matches = []
            self.indices = []

    @on(Input.Changed)
    def input_changed(self, input=Input()) -> None:
        if self._shell_tab_active():
            return  # No text to look for when inside shell

        keyword = input.value
        self._update_case_sensitivity()

        if not keyword or keyword != self.search_keyword:
            self._reset_search_state()

        if self.current_list_index is None and keyword:
            self._perform_search(keyword)

        if self.indices:
            self._update_input_display(keyword)
        else:
            self._show_no_results()

    @on(Input.Submitted)
    def input_submitted(self, input=Input()) -> None:
        if self._shell_tab_active():
            return  # No text to look for when inside shell

        keyword = input.value
        self._update_case_sensitivity()
        self._update_list_index()
        self._update_input_display(keyword)

    def _update_case_sensitivity(self) -> None:
        case_sensitive_switch = self.query_one("#case-sensitive-switch", Switch).value
        logs.case_sensitive = 0 if case_sensitive_switch else re.IGNORECASE

    def _reset_search_state(self) -> None:
        self.current_list_index = None
        self.search_keyword = ""
        self.indices = []

    def _perform_search(self, keyword: str) -> None:
        self.search_logs(keyword)
        self.search_keyword = keyword
        self.current_list_index = len(self.indices) - 1

    def _update_input_display(self, keyword: str) -> None:
        try:
            self.update_input_border()
        except IndexError:
            return  # No matches, don't update

        self.current_content_window.scroll_to(
            y=self._get_indices_list_value(), animate=False, duration=0
        )
        self.current_content_window.keyword = keyword
        self.current_content_window.current_index = self._get_indices_list_value()
        self.current_content_window.refresh()  # Force refresh, just to get rid of slow visual changes.

    def _show_no_results(self) -> None:
        input = self.query_one("#search_log_input")
        input.border_title = "No results"
        self.current_list_index = 0

    def _update_list_index(self) -> None:
        if self.current_list_index == 0:
            self.current_list_index = len(self.indices) - 1
        else:
            self.current_list_index -= 1

    def _get_indices_list_value(self):
        return self.indices[self.current_list_index]

    def update_input_border(self) -> None:
        line = str(self._get_indices_list_value())
        input = self.query_one("#search_log_input")
        input.border_title = f"{self.current_list_index + 1}/{len(self.indices)}"
        input.border_subtitle = f"line {line}"

    def live_logs_task(self):
        self.docker_manager.live_container_logs(logs, self.log_task_stop_event)

    def run_log_task(self):
        self.log_task_stop_event.set()  # Signal any existing task to stop
        self.log_task_stop_event = Event()  # Create a new stop event for the new task
        Thread(target=self.live_logs_task, daemon=True).start()

    def live_statistics_task(self):
        stop_event = Event()

        while not stop_event.is_set():
            try:
                stats = self._fetch_container_stats()
                cpu, memory = self._parse_stats(stats)
                self._update_logs(cpu, memory)
            except Exception:
                stop_event.set()
                self.live_logs_task()

    def _fetch_container_stats(self):
        return self.docker_manager.selected_container.stats(stream=False)

    def _parse_stats(self, stats: dict[Any, Any]):
        try:
            cpu_usage = (
                stats["cpu_stats"]["cpu_usage"]["total_usage"]
                / stats["cpu_stats"]["system_cpu_usage"]
            )
            memory_usage_mb = stats["memory_stats"]["usage"] / (1024 * 1024)
            cpu = f"{cpu_usage * 100:.3f}%"
            memory = f"{memory_usage_mb:.3f} MB"
            return cpu, memory
        except KeyError:
            # TODO: Correctly handle this case.
            return "N/A", "N/A"

    def _update_logs(self, cpu: str, memory: str):
        logs.border_subtitle = f"cpu: {cpu} | ram: {memory} | logs: {len(logs.lines)}"

    def _shell_tab_active(self):
        return self.query_one(TabbedContent).active == "shellpane"



from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label


class TopBar(Container):
    HELP_TEXT = "[b yellow]?[/b yellow] Help  [b yellow]p[/b yellow] Settings "

    def __init__(self, app_version: str, statuses=""):
        super().__init__()

        self.topbar_title = Label(
            f" [b]Pocker[/b] v{app_version}",
            id="topbar_title",
        )
        self.topbar_statuses = Label(statuses, id="topbar_statuses")
        self.topbar_help = Label(self.HELP_TEXT, id="topbar_help")

    def compose(self) -> ComposeResult:
        yield self.topbar_title
        yield self.topbar_statuses
        yield self.topbar_help



from pydantic import BaseModel
from yaml import dump, safe_load

from application.util.helper import POCKER_CONFIG_BASE_PATH

CONFIG_PATH = POCKER_CONFIG_BASE_PATH / "config.yaml"


class Config(BaseModel):
    log_tail: int = 2000
    max_log_lines: int = 2000
    start_wrap: bool = False
    show_all_containers: bool = False
    start_fullscreen: bool = False
    start_scroll: bool = True
    keymap: dict[str, str] = {
        "quit": "q",
        "logs": "l",
        "attributes": "a",
        "environment": "e",
        "statistics": "d",
        "shell": "v",
        "fullscreen": "f",
        "fullscreen-ci": "n",
        "wrap-logs": "w",
        "toggle-scroll": "s",
    }

    def __repr__(self) -> str:
        return super().__repr__(self.start_wrap)

    def create_default_config(self):
        default_config = {
            "log_tail": self.log_tail,
            "max_log_lines": self.max_log_lines,
            "start_wrap": self.start_wrap,
            "show_all_containers": self.show_all_containers,
            "start_fullscreen": self.start_fullscreen,
            "start_scroll": self.start_scroll,
            "keymap": self.keymap,
        }
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        with CONFIG_PATH.open("w") as file:
            dump(default_config, file)

    def save(self):
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        with CONFIG_PATH.open("w") as file:
            dump(self.model_dump(), file)


def load_config():
    if not CONFIG_PATH.exists():
        Config().create_default_config()
    with CONFIG_PATH.open() as file:
        config_dict = safe_load(file)
        return Config(**config_dict)
